// Copyright IBM Corp. 2016. All Rights Reserved.
// Node module: apiconnect-cli-apis
// US Government Users Restricted Rights - Use, duplication or disclosure
// restricted by GSA ADP Schedule Contract with IBM Corp.

/********************************************************* {COPYRIGHT-TOP} ***
 * IBM Confidential
 * OCO Source Materials
 * 5725-C11, 5725-C12, 7198-8FX, 5725-F70, 5725-F71, 5725-F72, 5725-L30
 *
 * (C) Copyright IBM Corporation 2015  All Rights Reserved.
 *
 * The source code for this program is not published or otherwise
 * divested of its trade secrets, irrespective of what has been
 * deposited with the U.S. Copyright Office.
 ********************************************************* {COPYRIGHT-END} **/

var ApicCliUtil = require('apiconnect-cli-util');
var ServerError = ApicCliUtil.ServerError;

var _ = require('lodash');
var async = require('async');
var apicUtil = ApicCliUtil.Util;
var consts = require('./consts');
var context = ApicCliUtil.Context;
var fs = require('fs');
var providerOrgName = ApicCliUtil.Util.providerOrgName;
var request = ApicCliUtil.Request;
var requestDefaultOptions = ApicCliUtil.Util.requestDefaultOptions;
var util = require('./util/util');
var yaml = require('js-yaml');
//var SharedCtx = require('./util/shared_ctx');
//var Project = ApicCliUtil.Project;
//var catalogName = ApicCliUtil.Util.catalogName;
//var Catalogs = ApicCliUtil.Catalogs;

/**
 * Default constructor
 *
 * @constructor
 */
function APIs() {}

APIs.prototype.listAPIs = function(options, callback) {
  var self = this;
  context.set(options);
  context.createCatalogContext(function(err, ctx) {
    if (err) { return callback(err); }
    self.fetchCatalogAPIs(ctx, function(err, rsp) {
      return callback(err, rsp);
    });
  });
};

APIs.prototype.fetchCatalogAPIs = function(catalogs, callback) {
  var self = this;
  var tasks = [];
  tasks.push(function(cb) {
    cb(null, {});
  });
  var mycatalogs = catalogs;
  if (typeof catalogs === 'object') { //not an array
    mycatalogs = [];
    _.keys(catalogs).forEach(function(orgName) {
      catalogs[orgName].forEach(function(catalog) {
        catalog.orgName = orgName;
        mycatalogs.push(catalog);
      });
    });
  }
  mycatalogs.forEach(function(catalog) {
    var t = (function(catalog) {
      return {
        invoke: function(rsp, cb) {
          self.fetchAPIs(catalog.id, catalog.orgName, function(err, result) {
            rsp[catalog.id] = {};
            rsp[catalog.id].apis = result;
            rsp[catalog.id].catalog = catalog;
            cb(err, rsp);
          });
        },
      };
    })(catalog);
    tasks.push(t.invoke);
  });

  async.waterfall(tasks, function(err, rsp) {
    if (err) {
      return callback(err, null);
    } else {
      return callback(null, rsp);
    }
  });
};

APIs.prototype.fetchAPIs = function(envNameOrId, orgName, callback) {
  var reqOpts = requestDefaultOptions();
  // /v1/orgs/:id/environments/:id/apis
  reqOpts.url += '/v1/orgs/' + orgName + '/environments/' + envNameOrId + '/apis';
  reqOpts.headers['accept'] = 'application/json';

  request.get(reqOpts, function(err, res, body) {
    if (err) {
      return callback(err, null);
    }
    if (res.statusCode !== 200) {
      err = new ServerError(res.statusCode, body);
      return callback(err, null);
    }
    var apis = typeof body === 'string' ? JSON.parse(body) : body;
    //filter the response
    return callback(null, apis);
  });
};

APIs.prototype.getAPI = function(options, apiVersion, callback) {
  var self = this;
  context.set(options);
  context.createCatalogContext(function(err, ctx) {
    if (err) { return callback(err); }
    self.fetchCatalogAPI(ctx, apiVersion, consts.HEADER_1, function(err, rsp) {
      return callback(err, rsp);
    });
  });
};

APIs.prototype.fetchCatalogAPI = function(catalogs, apiVersion, header, callback) {
  var self = this;
  var tasks = [];
  tasks.push(function(cb) {
    cb(null, {});
  });
  var mycatalogs = catalogs;
  if (typeof catalogs === 'object') { //not an array
    mycatalogs = [];
    _.keys(catalogs).forEach(function(orgName) {
      catalogs[orgName].forEach(function(catalog) {
        catalog.orgName = orgName;
        mycatalogs.push(catalog);
      });
    });
  }
  mycatalogs.forEach(function(catalog) {
    var t = (function(catalog, apiVersion, header) {
      return {
        invoke: function(rsp, cb) {
          self.fetchAPI(catalog.id, apiVersion, header, function(err, api) {
            rsp[catalog.id] = {};
            rsp[catalog.id].api = api;
            rsp[catalog.id].catalog = catalog;

            if (!api || api.length <= 0) {
              return cb(err, rsp);
            }

            //var apiId = api[0].id; // To do: Defect in the backend, only works with name:version and not id.
            var apiId = api[0].name + ':' + api[0].version;
            self.getRelatedProductsFromCatalog(apiId, catalog.id, function(err, relatedProducts) {
              if (err) { return cb(err, null); }
              rsp[catalog.id].relatedProducts = {};
              rsp[catalog.id].relatedProducts[apiId] = relatedProducts;
              cb(err, rsp);
            });
          });
        },
      };
    })(catalog, apiVersion, header);
    tasks.push(t.invoke);
  });

  async.waterfall(tasks, function(err, rsp) {
    if (err) {
      return callback(err, null);
    } else {
      return callback(null, rsp);
    }
  });
};

APIs.prototype.getRelatedProductsFromCatalog = function(apiId, catalogId, cb) {
  var reqOpts = requestDefaultOptions();
  // GET /v1/orgs/{orgId}/environments/{envId}/search?apiIdentifier=
  reqOpts.url += '/v1/orgs/' + providerOrgName() + '/environments/' + catalogId +
    '/search?apiIdentifier=' + apiId + '&type=productVersion';
  reqOpts.headers['accept'] = 'application/json'; //'application/vnd.ibm-apim.swagger2+json';
  request.get(reqOpts, function(err, res, body) {
    if (err) {
      return cb(err, null);
    }
    if (res.statusCode !== 200) {
      err = new ServerError(res.statusCode, body);
      return cb(err, null);
    }
    var relProds = typeof body === 'string' ? JSON.parse(body) : body;
    cb(null, relProds);
  });
};

APIs.prototype.fetchAPI = function(envNameOrId, apiVersion, header, callback) {
  var apiVersionUrlPart = util.getNameVersionUrlPart(apiVersion);

  var reqOpts = requestDefaultOptions();

  reqOpts.url += '/v1/orgs/' + providerOrgName() + '/environments/' + envNameOrId + '/search?' + apiVersionUrlPart;
//  reqOpts.headers['accept'] = header;

  request.get(reqOpts, function(err, res, body) {
    if (err) {
      return callback(err, null);
    }
    if (res.statusCode !== 200) {
      err = new ServerError(res.statusCode, body);
      return callback(err, null);
    }
    var results = typeof body === 'string' ? JSON.parse(body) : body;
    var apis = [];
    if (results) {
      for (var i = 0; i < results.length; i++) {
        // The search might return products as well as apis
        // The search might return multiple apis as well, need to find the right one
        if (results[i].type === consts.API_TYPE && results[i].name === _.split(apiVersion, ':')[0]) {
          apis.push(results[i]);
        }
      }
    }

    //filter the response
    return callback(null, apis);
  });
};

APIs.prototype.setAPI = function(options, apiVersion, status, callback) {
  var self = this;

  context.set(options);
  context.createCatalogContext(function(err, ctx) {
    if (err) { return callback(err); }
    self.fetchCatalogAPI(ctx, apiVersion, consts.HEADER_1, function(err, rsp) {
      if (err) {
        return callback(err, null);
      }

      var catalogId = _.keys(rsp)[0];
      var orgId = rsp[catalogId].catalog.orgName;

      if (rsp[catalogId].api && rsp[catalogId].api.length > 0) {
        var apiId = rsp[catalogId].api[0].id;

        stopOrStartAPI(catalogId, orgId, apiId, status, function(err, api) {
          return callback(err, api);
        });
      } else {
        return callback(err, null);
      }
    });
  });
};

function stopOrStartAPI(catalogId, orgId, apiId, status, callback) {
  var err;

  if (!status || (status !== 'stop' && status !== 'start')) {
    err = new Error('The --status option is required.');
    return callback(err, null);
  }

  var reqOpts = requestDefaultOptions();
  // https://localhost:8443/v1/orgs/tenantzero/environments/sb/apis/5645cc669b48b831e5318c83/stop
  reqOpts.url += '/v1/orgs/' + orgId + '/environments/' + catalogId + '/apis/' + apiId + '/' + status;
//  reqOpts.headers['accept'] = consts.HEADER_1;

  request.post(reqOpts, function(err, res, body) {
    if (err) {
      return callback(err, null);
    }
    if (res.statusCode !== 200) {
      err = new ServerError(res.statusCode, body);
      return callback(err, null);
    }
    var api = typeof body === 'string' ? JSON.parse(body) : body;
    return callback(null, api);
  });
}

//accept header (application/vnd.ibm-apim.swagger2+json) and GET /v1/orgs/:id/environments/:id/apis/:id
APIs.prototype.pullAPI = function(options, apiVersion, callback) {
  var self = this;
  context.set(options);

  async.waterfall([
    function(api_callback) {
      return context.createCatalogContext(function(err, ctx) {
        if (err) {
          return callback(err);
        }
        return api_callback(null, ctx);
      });
    },
    function(ctx, api_callback) {
      return self.fetchCatalogAPI(ctx, apiVersion, consts.HEADER_1, function(err, rsp) {
        if (err) {
          return callback(err);
        }
        return api_callback(null, rsp);
      });
    },
    function(rsp, api_callback) {
      var catalogId = _.keys(rsp)[0];
      var apiId = rsp[catalogId].api[0].id;
      return downloadApi(catalogId, apiId, function(err, result) {
        if (err) {
          return callback(err);
        }
        return api_callback(null, result);
      });
    },
    function(result, api_callback) {
      var api = result.api;
      if (api && api['x-ibm-configuration'] && api['x-ibm-configuration'].type === 'wsdl') {
        var wsdlFileName = api.info['x-ibm-name'] + '_' + api.info.version + '.wsdl';
        var id = api['x-ibm-configuration']['wsdl-definition'].wsdl;
        return wsdlDownloadReq(id, function(err, resp, body) {
          if (err) {
            return api_callback(err);
          }
          if (resp.headers['content-disposition'].split('.').pop().replace(/['"]+/g, '') === 'zip') {
            wsdlFileName = api.info['x-ibm-name'] + '_' + api.info.version + '.zip';
          }
          fs.writeFileSync(wsdlFileName, body);
          result.wsdl = wsdlFileName;

          return api_callback(null, result);
        });
      } else {
        return api_callback(null, result);
      }
    },
  ],
  function(err, result) {
    if (err) {
      return callback(err);
    }
    return callback(null, result);
  });
};

function downloadApi(catalogId, apiId, callback) {
  var reqOpts = requestDefaultOptions();
  reqOpts.url += '/v1/orgs/' + providerOrgName() + '/environments/' + catalogId + '/apis/' + apiId;
  delete reqOpts.headers.accept;

  request.get(reqOpts, function(err, res, body) {
    if (err) {
      return callback(err, null);
    }
    if (res.statusCode !== 200) {
      err = new ServerError(res.statusCode, body);
      return callback(err, null);
    }
    var api = typeof body === 'string' ? JSON.parse(body) : body;
    //Todo: below is hack till backend fixes the swagger
    //var filename = api.info['x-ibm-name'] + '_' + api.info.version + '.yaml';
    //var nameVersion = api.info['x-ibm-name'] + ':' + api.info.version;
    var filename = api.info['x-ibm-name'] + '_' + api.info.version + '.yaml';
    var nameVersion = api.info['x-ibm-name'] + ':' + api.info.version;

    util.validateWriteFile(filename, yaml.dump(api), true);

    var result = { nameVersion: nameVersion, filename: filename, api: api };
    return callback(null, result);
  });
}

function wsdlDownloadReq(id, cb) {
  var opts = apicUtil.requestDefaultOptions();
  opts.url += '/v1/orgs/' + providerOrgName() + '/apis/wsdl/' + id;
  opts.headers['accept'] = 'application/json';
  opts.encoding = null;

  return request.get(opts, function(err, resp, body) {
    if (err) {
      return cb(err);
    }
    if (resp.statusCode < 200 || resp.statusCode > 399) {
      return cb(new ServerError(resp, body));
    }
    return cb(null, resp, body);
  });
}

APIs.prototype.cloneAPI = function(options, callback) {
  var self = this;
  var clonedApis = {};
  var tasks = [];
  tasks.push(function(cb) {
    cb(null, {});
  });

  context.set(options);
  context.createCatalogContext(function(err, ctx) {
    if (err) { return callback(err); }
    self.fetchCatalogAPIs(ctx, function(err, apis) {
      if (err) {
        return callback(err, null);
      }

      _.keys(apis).forEach(function(key) {
        clonedApis[key] = {};
        clonedApis[key].catalogShortName = apis[key].catalog.shortName;
        clonedApis[key].orgName = apis[key].catalog.orgName;
        clonedApis[key].apis = [];
        var i = 0;
        apis[key].apis.forEach(function(api) {
          var t1 = (function(key, api, clonedApis) {
            return {
              invoke: function(rsp, cb) {
                var catalogId = key;
                var apiId = api.id;

                async.waterfall([
                  function(api_callback) {
                    return downloadApi(catalogId, apiId, api_callback);
                  },
                  function(result, api_callback) {
                    var api_swagger = result.api;
                    if (api_swagger && api_swagger['x-ibm-configuration'] &&
                        api_swagger['x-ibm-configuration'].type === 'wsdl') {
                      var wsdlFileName = api_swagger.info['x-ibm-name'] + '_' + api_swagger.info.version + '.wsdl';
                      var id = api_swagger['x-ibm-configuration']['wsdl-definition'].wsdl;
                      return wsdlDownloadReq(id, function(err, resp, body) {
                        if (err) {
                          return api_callback(err);
                        }
                        if (resp.headers['content-disposition'].split('.').pop().replace(/['"]+/g, '') === 'zip') {
                          wsdlFileName = api_swagger.info['x-ibm-name'] + '_' + api_swagger.info.version + '.zip';
                        }
                        return fs.writeFile(wsdlFileName, body, function(err) {
                          if (err) {
                            return api_callback(err);
                          }
                          clonedApis[key].apis[i] = {};
                          clonedApis[key].apis[i].nameVersion = result.nameVersion;
                          clonedApis[key].apis[i].filename = result.filename;
                          clonedApis[key].apis[i].wsdl = wsdlFileName;
                          i++;

                          return api_callback(null, clonedApis);
                        });
                      });
                    } else {
                      clonedApis[key].apis[i] = {};
                      clonedApis[key].apis[i].nameVersion = result.nameVersion;
                      clonedApis[key].apis[i].filename = result.filename;
                      i++;
                      return api_callback(null, clonedApis);
                    }
                  },
                ],
                function(err, clonedApis) {
                  if (err) {
                    return cb(err);
                  }
                  return cb(null, clonedApis);
                });
              },
            };
          })(key, api, clonedApis);
          tasks.push(t1.invoke);
        });

        async.waterfall(tasks, callback);
      });
    });
  });
};

module.exports = APIs;
